@@ -4,11 +4,14 @@ import (
 	"bytes"
 	"errors"
 	"fmt"
+	"io"
+	"io/ioutil"
 	"math"
 	"net/http"
 	"net/http/httptest"
 	"net/url"
 	"reflect"
+	"strings"
 	"testing"
 )
 
@@ -48,15 +51,17 @@ func TestNew(t *testing.T) {
 }
 
 func TestSlingNew(t *testing.T) {
+	fakeBodyProvider := jsonBodyProvider{FakeModel{}}
+
 	cases := []*Sling{
 		&Sling{httpClient: &http.Client{}, method: "GET", rawURL: "http://example.com"},
 		&Sling{httpClient: nil, method: "", rawURL: "http://example.com"},
 		&Sling{queryStructs: make([]interface{}, 0)},
 		&Sling{queryStructs: []interface{}{paramsA}},
 		&Sling{queryStructs: []interface{}{paramsA, paramsB}},
-		&Sling{bodyJSON: &FakeModel{Text: "a"}},
-		&Sling{bodyJSON: FakeModel{Text: "a"}},
-		&Sling{bodyJSON: nil},
+		&Sling{bodyProvider: fakeBodyProvider},
+		&Sling{bodyProvider: fakeBodyProvider},
+		&Sling{bodyProvider: nil},
 		New().Add("Content-Type", "application/json"),
 		New().Add("A", "B").Add("a", "c").New(),
 		New().Add("A", "B").New().Add("a", "c"),
@@ -66,7 +71,7 @@ func TestSlingNew(t *testing.T) {
 	for _, sling := range cases {
 		child := sling.New()
 		if child.httpClient != sling.httpClient {
-			t.Errorf("expected %p, got %p", sling.httpClient, child.httpClient)
+			t.Errorf("expected %v, got %v", sling.httpClient, child.httpClient)
 		}
 		if child.method != sling.method {
 			t.Errorf("expected %s, got %s", sling.method, child.method)
@@ -94,13 +99,9 @@ func TestSlingNew(t *testing.T) {
 				t.Errorf("child.queryStructs was a re-slice, expected slice with copied contents")
 			}
 		}
-		// bodyJSON should be copied
-		if child.bodyJSON != sling.bodyJSON {
-			t.Errorf("expected %v, got %v", sling.bodyJSON, child.bodyJSON)
-		}
-		// bodyForm should be copied
-		if child.bodyForm != sling.bodyForm {
-			t.Errorf("expected %v, got %v", sling.bodyForm, child.bodyForm)
+		// body should be copied
+		if child.bodyProvider != sling.bodyProvider {
+			t.Errorf("expected %v, got %v", sling.bodyProvider, child.bodyProvider)
 		}
 	}
 }
@@ -118,7 +119,25 @@ func TestClientSetter(t *testing.T) {
 		sling := New()
 		sling.Client(c.input)
 		if sling.httpClient != c.expected {
-			t.Errorf("expected %v, got %v", c.expected, sling.httpClient)
+			t.Errorf("input %v, expected %v, got %v", c.input, c.expected, sling.httpClient)
+		}
+	}
+}
+
+func TestDoerSetter(t *testing.T) {
+	developerClient := &http.Client{}
+	cases := []struct {
+		input    Doer
+		expected Doer
+	}{
+		{nil, http.DefaultClient},
+		{developerClient, developerClient},
+	}
+	for _, c := range cases {
+		sling := New()
+		sling.Doer(c.input)
+		if sling.httpClient != c.expected {
+			t.Errorf("input %v, expected %v, got %v", c.input, c.expected, sling.httpClient)
 		}
 	}
 }
@@ -229,6 +248,34 @@ func TestSetHeader(t *testing.T) {
 	}
 }
 
+func TestBasicAuth(t *testing.T) {
+	cases := []struct {
+		sling        *Sling
+		expectedAuth []string
+	}{
+		// basic auth: username & password
+		{New().SetBasicAuth("Aladdin", "open sesame"), []string{"Aladdin", "open sesame"}},
+		// empty username
+		{New().SetBasicAuth("", "secret"), []string{"", "secret"}},
+		// empty password
+		{New().SetBasicAuth("admin", ""), []string{"admin", ""}},
+	}
+	for _, c := range cases {
+		req, err := c.sling.Request()
+		if err != nil {
+			t.Errorf("unexpected error when building Request with .SetBasicAuth()")
+		}
+		username, password, ok := req.BasicAuth()
+		if !ok {
+			t.Errorf("basic auth missing when expected")
+		}
+		auth := []string{username, password}
+		if !reflect.DeepEqual(c.expectedAuth, auth) {
+			t.Errorf("not DeepEqual: expected %v, got %v", c.expectedAuth, auth)
+		}
+	}
+}
+
 func TestQueryStructSetter(t *testing.T) {
 	cases := []struct {
 		sling           *Sling
@@ -261,24 +308,26 @@ func TestQueryStructSetter(t *testing.T) {
 
 func TestBodyJSONSetter(t *testing.T) {
 	fakeModel := &FakeModel{}
+	fakeBodyProvider := jsonBodyProvider{payload: fakeModel}
+
 	cases := []struct {
-		initial  interface{}
+		initial  BodyProvider
 		input    interface{}
-		expected interface{}
+		expected BodyProvider
 	}{
 		// json tagged struct is set as bodyJSON
-		{nil, fakeModel, fakeModel},
+		{nil, fakeModel, fakeBodyProvider},
 		// nil argument to bodyJSON does not replace existing bodyJSON
-		{fakeModel, nil, fakeModel},
+		{fakeBodyProvider, nil, fakeBodyProvider},
 		// nil bodyJSON remains nil
 		{nil, nil, nil},
 	}
 	for _, c := range cases {
 		sling := New()
-		sling.bodyJSON = c.initial
+		sling.bodyProvider = c.initial
 		sling.BodyJSON(c.input)
-		if sling.bodyJSON != c.expected {
-			t.Errorf("expected %v, got %v", c.expected, sling.bodyJSON)
+		if sling.bodyProvider != c.expected {
+			t.Errorf("expected %v, got %v", c.expected, sling.bodyProvider)
 		}
 		// Header Content-Type should be application/json if bodyJSON arg was non-nil
 		if c.input != nil && sling.header.Get(contentType) != jsonContentType {
@@ -290,24 +339,27 @@ func TestBodyJSONSetter(t *testing.T) {
 }
 
 func TestBodyFormSetter(t *testing.T) {
+	fakeParams := FakeParams{KindName: "recent", Count: 25}
+	fakeBodyProvider := formBodyProvider{payload: fakeParams}
+
 	cases := []struct {
-		initial  interface{}
+		initial  BodyProvider
 		input    interface{}
-		expected interface{}
+		expected BodyProvider
 	}{
 		// url tagged struct is set as bodyStruct
-		{nil, paramsB, paramsB},
+		{nil, paramsB, fakeBodyProvider},
 		// nil argument to bodyStruct does not replace existing bodyStruct
-		{paramsB, nil, paramsB},
+		{fakeBodyProvider, nil, fakeBodyProvider},
 		// nil bodyStruct remains nil
 		{nil, nil, nil},
 	}
 	for _, c := range cases {
 		sling := New()
-		sling.bodyForm = c.initial
+		sling.bodyProvider = c.initial
 		sling.BodyForm(c.input)
-		if sling.bodyForm != c.expected {
-			t.Errorf("expected %v, got %v", c.expected, sling.bodyForm)
+		if sling.bodyProvider != c.expected {
+			t.Errorf("expected %v, got %v", c.expected, sling.bodyProvider)
 		}
 		// Content-Type should be application/x-www-form-urlencoded if bodyStruct was non-nil
 		if c.input != nil && sling.header.Get(contentType) != formContentType {
@@ -316,7 +368,32 @@ func TestBodyFormSetter(t *testing.T) {
 			t.Errorf("did not expect a Content-Type header, got %s", sling.header.Get(contentType))
 		}
 	}
+}
 
+func TestBodySetter(t *testing.T) {
+	fakeInput := ioutil.NopCloser(strings.NewReader("test"))
+	fakeBodyProvider := bodyProvider{body: fakeInput}
+
+	cases := []struct {
+		initial  BodyProvider
+		input    io.Reader
+		expected BodyProvider
+	}{
+		// nil body is overriden by a set body
+		{nil, fakeInput, fakeBodyProvider},
+		// initial body is not overriden by nil body
+		{fakeBodyProvider, nil, fakeBodyProvider},
+		// nil body is returned unaltered
+		{nil, nil, nil},
+	}
+	for _, c := range cases {
+		sling := New()
+		sling.bodyProvider = c.initial
+		sling.Body(c.input)
+		if sling.bodyProvider != c.expected {
+			t.Errorf("expected %v, got %v", c.expected, sling.bodyProvider)
+		}
+	}
 }
 
 func TestRequest_urlAndMethod(t *testing.T) {
@@ -403,6 +480,9 @@ func TestRequest_body(t *testing.T) {
 		{New().BodyJSON(modelA).New().BodyForm(paramsB), "count=25&kind_name=recent", formContentType},
 		{New().BodyForm(paramsB).New().BodyJSON(nil), "count=25&kind_name=recent", formContentType},
 		{New().BodyJSON(modelA).New().BodyForm(nil), "{\"text\":\"note\",\"favorite_count\":12}\n", jsonContentType},
+		// Body
+		{New().Body(strings.NewReader("this-is-a-test")), "this-is-a-test", ""},
+		{New().Body(strings.NewReader("a")).Body(strings.NewReader("b")), "b", ""},
 	}
 	for _, c := range cases {
 		req, _ := c.sling.Request()
@@ -412,8 +492,8 @@ func TestRequest_body(t *testing.T) {
 		if value := buf.String(); value != c.expectedBody {
 			t.Errorf("expected Request.Body %s, got %s", c.expectedBody, value)
 		}
-		// Header Content-Type should be application/json
-		if actualHeader := req.Header.Get(contentType); actualHeader != c.expectedContentType {
+		// Header Content-Type should be expectedContentType ("" means no contentType expected)
+		if actualHeader := req.Header.Get(contentType); actualHeader != c.expectedContentType && c.expectedContentType != "" {
 			t.Errorf("Incorrect or missing header, expected %s, got %s", c.expectedContentType, actualHeader)
 		}
 	}
@@ -634,26 +714,6 @@ func TestDo_onFailureWithNilValue(t *testing.T) {
 	}
 }
 
-func TestDo_skipDecodingIfContentTypeWrong(t *testing.T) {
-	client, mux, server := testServer()
-	defer server.Close()
-	mux.HandleFunc("/success", func(w http.ResponseWriter, r *http.Request) {
-		w.Header().Set("Content-Type", "text/html")
-		fmt.Fprintf(w, `{"text": "Some text", "favorite_count": 24}`)
-	})
-
-	sling := New().Client(client)
-	req, _ := http.NewRequest("GET", "http://example.com/success", nil)
-
-	model := new(FakeModel)
-	sling.Do(req, model, nil)
-
-	expectedModel := &FakeModel{}
-	if !reflect.DeepEqual(expectedModel, model) {
-		t.Errorf("decoding should have been skipped, Content-Type was incorrect")
-	}
-}
-
 func TestReceive_success(t *testing.T) {
 	client, mux, server := testServer()
 	defer server.Close()
